import { ScheduledPayloadSchema, currentDate, } from "@systemfsoftware/trigger.dev_core";
import cronstrue from "cronstrue";
import { runLocalStorage } from "../runLocalStorage.js";
const examples = [
    {
        id: "now",
        name: "Now",
        icon: "clock",
        payload: {
            ts: currentDate.marker,
            lastTimestamp: currentDate.marker,
        },
    },
];
export class IntervalTrigger {
    options;
    constructor(options) {
        this.options = options;
    }
    get event() {
        return {
            name: "trigger.scheduled",
            title: "Schedule",
            source: "trigger.dev",
            icon: "schedule-interval",
            examples,
            parsePayload: ScheduledPayloadSchema.parse,
            properties: [
                {
                    label: "Interval",
                    text: `${this.options.seconds}s`,
                },
            ],
        };
    }
    attachToJob(triggerClient, job) { }
    get preprocessRuns() {
        return false;
    }
    async verifyPayload(payload) {
        return { success: true };
    }
    toJSON() {
        return {
            type: "scheduled",
            schedule: {
                type: "interval",
                options: {
                    seconds: this.options.seconds,
                },
            },
        };
    }
}
/** `intervalTrigger()` is set as a [Job's trigger](/sdk/job) to trigger a Job at a recurring interval.
 * @param options An object containing options about the interval.
 */
export function intervalTrigger(options) {
    return new IntervalTrigger(options);
}
export class CronTrigger {
    #options;
    constructor(options) {
        this.#options = options;
    }
    get event() {
        /**
         * We need to concat `(UTC)` string at the end of the human readable string to avoid confusion
         * with execution time/last run of a job in the UI dashboard which is displayed in local time.
         */
        const humanReadable = cronstrue.toString(this.#options.cron, {
            throwExceptionOnParseError: false,
        }).concat(" (UTC)");
        return {
            name: "trigger.scheduled",
            title: "Cron Schedule",
            source: "trigger.dev",
            icon: "schedule-cron",
            examples,
            parsePayload: ScheduledPayloadSchema.parse,
            properties: [
                {
                    label: "cron",
                    text: this.#options.cron,
                },
                {
                    label: "Schedule",
                    text: humanReadable,
                },
            ],
        };
    }
    attachToJob(triggerClient, job) { }
    get preprocessRuns() {
        return false;
    }
    async verifyPayload(payload) {
        return { success: true };
    }
    toJSON() {
        return {
            type: "scheduled",
            schedule: {
                type: "cron",
                options: {
                    cron: this.#options.cron,
                },
            },
        };
    }
}
/** `cronTrigger()` is set as a [Job's trigger](https://trigger.dev/docs/sdk/job) to trigger a Job on a recurring schedule using a CRON expression.
 * @param options An object containing options about the CRON schedule.
 */
export function cronTrigger(options) {
    return new CronTrigger(options);
}
/** DynamicSchedule` allows you to define a scheduled trigger that can be configured dynamically at runtime. */
export class DynamicSchedule {
    client;
    options;
    /**
     * @param client The `TriggerClient` instance to use for registering the trigger.
     * @param options The options for the schedule.
     */
    constructor(client, options) {
        this.client = client;
        this.options = options;
        client.attachDynamicSchedule(this.options.id);
    }
    get id() {
        return this.options.id;
    }
    get event() {
        return {
            name: "trigger.scheduled",
            title: "Dynamic Schedule",
            source: "trigger.dev",
            icon: "schedule-dynamic",
            examples,
            parsePayload: ScheduledPayloadSchema.parse,
        };
    }
    async register(key, metadata) {
        const runStore = runLocalStorage.getStore();
        if (!runStore) {
            return this.client.registerSchedule(this.id, key, metadata);
        }
        const { io } = runStore;
        return await io.runTask([key, "register"], async (task) => {
            return this.client.registerSchedule(this.id, key, metadata);
        }, {
            name: "Register Schedule",
            icon: metadata.type === "cron" ? "schedule-cron" : "schedule-interval",
            properties: [
                { label: "Dynamic Schedule", text: this.id },
                { label: "Schedule ID", text: key },
            ],
            params: metadata,
        });
    }
    async unregister(key) {
        const runStore = runLocalStorage.getStore();
        if (!runStore) {
            return this.client.unregisterSchedule(this.id, key);
        }
        const { io } = runStore;
        return await io.runTask([key, "unregister"], async (task) => {
            return this.client.unregisterSchedule(this.id, key);
        }, {
            name: "Unregister Schedule",
            icon: "schedule",
            properties: [
                { label: "Dynamic Schedule", text: this.id },
                { label: "Schedule ID", text: key },
            ],
        });
    }
    attachToJob(triggerClient, job) {
        triggerClient.attachDynamicScheduleToJob(this.options.id, job);
    }
    get preprocessRuns() {
        return false;
    }
    async verifyPayload(payload) {
        return { success: true };
    }
    toJSON() {
        return {
            type: "dynamic",
            id: this.options.id,
        };
    }
}
//# sourceMappingURL=scheduled.js.map